from abc import ABC


class BaseStreamState(ABC):
    """Filler class in case we want to add abstract methods in the future,
    such as maybe requiring a method to clear/reset/copy StreamStates..."""
    pass
